# distutils: language = c++
import numpy as np
cimport numpy as np
# cimport csvd
cimport eg


cdef np.ndarray Vector3dToNumpy(eg.VectorXd cx):
    print('3')
    result = np.ndarray((cx.rows()))
    for i in range(cx.rows()):
        result[i] = cx[i]

    print('4')
    return result

cdef eg.VectorXd NumpyToVector3d (np.ndarray[double, ndim=1, mode="c"] x):
    print('1')
    cdef eg.VectorXd cx = eg.VectorXd(3)
    print('2')
    for i in range(3):
        print('i: ', i)
        cx[i] = x[i]
    print('3')
    return cx

cpdef xyz(i):
    return Vector3dToNumpy(NumpyToVector3d(i))

def abc(s):
    print(s)
'''
cdef np.ndarray Matrix3dToNumpy(Matrix3d m):
    result = np.ndarray([3, 3])
    for i in range(3):
        for j in range(3):
            result[i, j] = m.coeff(i, j)
    return result




cdef ensure_positive_Z(np.ndarray u, np.ndarray, v):
    Mat3d  = numpy to eigen
    eigne_result = csvd.ensure_positive_Z(csvd.Mat3d)
    resut = eigen to numpy(eigen_result)
    return result
'''



from sympy import symbols, simplify
from sympy import eye, Matrix, ZeroMatrix
import yaml


from se3 import getr, SXForm, Ex, Ey, perfect_Ez_pi

q_sym = symbols('q0:12')

x0, y0, z0 = symbols('x0, y0, z0', real=True)
x1, y1, z1 = symbols('x1, y1, z1', real=True)
x2, y2, z2 = symbols('x2, y2, z2', real=True)
x3, y3, z3 = symbols('x3, y3, z3', real=True)

x6, y6, z6 = symbols('x6, y6, z6', real=True)
x7, y7, z7 = symbols('x7, y7, z7', real=True)
x8, y8, z8 = symbols('x8, y8, z8', real=True)

x9, y9, z9 = symbols('x9, y9, z9', real=True)
x10, y10, z10 = symbols('x10, y10, z10', real=True)
x11, y11, z11 = symbols('x11, y11, z11', real=True)

x12, y12, z12 = symbols('x12, y12, z12', real=True)
x13, y13, z13 = symbols('x13, y13, z13', real=True)
x14, y14, z14 = symbols('x14, y14, z14', real=True)

x15, y15, z15 = symbols('x15, y15, z15', real=True)
x16, y16, z16 = symbols('x16, y16, z16', real=True)
x17, y17, z17 = symbols('x17, y17, z17', real=True)


r0 = Matrix([x0, y0, z0])
r1 = Matrix([x1, y1, z1])
r2 = Matrix([x2, y2, z2])
r3 = Matrix([x3, y3, z3])

r6 = Matrix([x6, y6, z6])
r7 = Matrix([x7, y7, z7])
r8 = Matrix([x8, y8, z8])
r9 = Matrix([x9, y9, z9])
r10 = Matrix([x10, y10, z10])
r11 = Matrix([x11, y11, z11])
r12 = Matrix([x12, y12, z12])
r13 = Matrix([x13, y13, z13])

r14 = Matrix([x14, y14, z14])
r15 = Matrix([x15, y15, z15])
r16 = Matrix([x16, y16, z16])
r17 = Matrix([x17, y17, z17])

XJ = [SXForm(Ex(q_sym[0]), ZeroMatrix(3, 1)),
      SXForm(Ey(q_sym[1]), ZeroMatrix(3, 1)),
      SXForm(Ey(q_sym[2]), ZeroMatrix(3, 1)),
      SXForm(Ex(q_sym[3]), ZeroMatrix(3, 1)),
      SXForm(Ey(q_sym[4]), ZeroMatrix(3, 1)),
      SXForm(Ey(q_sym[5]), ZeroMatrix(3, 1)),
      SXForm(Ex(q_sym[6]), ZeroMatrix(3, 1)),
      SXForm(Ey(q_sym[7]), ZeroMatrix(3, 1)),
      SXForm(Ey(q_sym[8]), ZeroMatrix(3, 1)),
      SXForm(Ex(q_sym[9]), ZeroMatrix(3, 1)),
      SXForm(Ey(q_sym[10]), ZeroMatrix(3, 1)),
      SXForm(Ey(q_sym[11]), ZeroMatrix(3, 1))]

Xtree = [None,  # 0
         None,  # 1
         None,  # 2
         None,  # 3
         None,  # 4
         None,  # 5
         SXForm(eye(3), r6),
         SXForm(perfect_Ez_pi, r7),
         SXForm(eye(3), r8),
         SXForm(eye(3), r9),
         SXForm(perfect_Ez_pi, r10),
         SXForm(eye(3), r11),
         SXForm(eye(3), r12),
         SXForm(perfect_Ez_pi, r13),
         SXForm(eye(3), r14),
         SXForm(eye(3), r15),
         SXForm(perfect_Ez_pi, r16),
         SXForm(eye(3), r17)]

gcLocation = [SXForm(eye(3), r0),
              SXForm(eye(3), r1),
              SXForm(eye(3), r2),
              SXForm(eye(3), r3)]

Abad0_SX = Xtree[6]
Abad1_SX = Xtree[9]
Abad2_SX = Xtree[12]
Abad3_SX = Xtree[15]

Abad0_SX = simplify(Abad0_SX)
Abad1_SX = simplify(Abad1_SX)
Abad2_SX = simplify(Abad2_SX)
Abad3_SX = simplify(Abad3_SX)

abad0 = getr(Abad0_SX)
abad1 = getr(Abad1_SX)
abad2 = getr(Abad2_SX)
abad3 = getr(Abad3_SX)

abad0 = simplify(abad0)
abad1 = simplify(abad1)
abad2 = simplify(abad2)
abad3 = simplify(abad3)


Hip0_SX = Xtree[7] * XJ[0] * Abad0_SX
Hip1_SX = Xtree[10] * XJ[3] * Abad1_SX
Hip2_SX = Xtree[13] * XJ[6] * Abad2_SX
Hip3_SX = Xtree[16] * XJ[9] * Abad3_SX

Hip0_SX = simplify(Hip0_SX)
Hip1_SX = simplify(Hip1_SX)
Hip2_SX = simplify(Hip2_SX)
Hip3_SX = simplify(Hip3_SX)

hip0 = getr(Hip0_SX)
hip1 = getr(Hip1_SX)
hip2 = getr(Hip2_SX)
hip3 = getr(Hip3_SX)

hip0 = simplify(hip0)
hip1 = simplify(hip1)
hip2 = simplify(hip2)
hip3 = simplify(hip3)


Knee0_SX = Xtree[8] * XJ[1] * Hip0_SX
Knee1_SX = Xtree[11] * XJ[4] * Hip1_SX
Knee2_SX = Xtree[14] * XJ[7] * Hip2_SX
Knee3_SX = Xtree[17] * XJ[10] * Hip3_SX

Knee0_SX = simplify(Knee0_SX)
Knee1_SX = simplify(Knee1_SX)
Knee2_SX = simplify(Knee2_SX)
Knee3_SX = simplify(Knee3_SX)

knee0 = getr(Knee0_SX)
knee1 = getr(Knee1_SX)
knee2 = getr(Knee2_SX)
knee3 = getr(Knee3_SX)

knee0 = simplify(knee0)
knee1 = simplify(knee1)
knee2 = simplify(knee2)
knee3 = simplify(knee3)

T0 = gcLocation[0] * XJ[2] * Knee0_SX
T1 = gcLocation[1] * XJ[5] * Knee1_SX
T2 = gcLocation[2] * XJ[8] * Knee2_SX
T3 = gcLocation[3] * XJ[11] * Knee3_SX

T0 = simplify(T0)
T1 = simplify(T1)
T2 = simplify(T2)
T3 = simplify(T3)

foot0 = getr(T0)
foot1 = getr(T1)
foot2 = getr(T2)
foot3 = getr(T3)

foot0 = simplify(foot0)
foot1 = simplify(foot1)
foot2 = simplify(foot2)
foot3 = simplify(foot3)


def links_reader(file):

    with open(file, 'r') as stream:
        try:
            out = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)

    links_params = {x6: out['abad_location'][0],
                    y6: -out['abad_location'][1],
                    z6: out['abad_location'][2],
                    x7: out['hip_location'][0],
                    y7: -out['hip_location'][1],
                    z7: out['hip_location'][2],
                    x8: out['knee_location'][0],
                    y8: out['knee_location'][1],
                    z8: out['knee_location'][2],
                    x0: 0,
                    y0: 0,
                    z0: -out['knee_link_length'],
                    # Leg 1
                    x9: out['abad_location'][0],
                    y9: out['abad_location'][1],
                    z9: out['abad_location'][2],
                    x10: out['hip_location'][0],
                    y10: out['hip_location'][1],
                    z10: out['hip_location'][2],
                    x11: out['knee_location'][0],
                    y11: out['knee_location'][1],
                    z11: out['knee_location'][2],
                    x1: 0,
                    y1: 0,
                    z1: -out['knee_link_length'],
                    # Leg 2
                    x12: -out['abad_location'][0],
                    y12: -out['abad_location'][1],
                    z12: out['abad_location'][2],
                    x13: out['hip_location'][0],
                    y13: -out['hip_location'][1],
                    z13: out['hip_location'][2],
                    x14: out['knee_location'][0],
                    y14: out['knee_location'][1],
                    z14: out['knee_location'][2],
                    x2: 0,
                    y2: 0,
                    z2: -out['knee_link_length'],
                    # Leg 3
                    x15: -out['abad_location'][0],
                    y15: out['abad_location'][1],
                    z15: out['abad_location'][2],
                    x16: out['hip_location'][0],
                    y16: out['hip_location'][1],
                    z16: out['hip_location'][2],
                    x17: out['knee_location'][0],
                    y17: out['knee_location'][1],
                    z17: out['knee_location'][2],
                    x3: 0,
                    y3: 0,
                    z3: -out['knee_link_length']}

    return links_params



# cdef extern from "<Eigen/Dense>" namespace "Eigen":
cdef extern from "../include/cppTypes.h":
    cdef cppclass Matrix[T, U, V]:
        Matrix()
        T& operator()(int, int)
        T& operator()(int)

    # A matrix
    cdef cppclass MatrixXd:
        MatrixXd()
        MatrixXd(int, int)
        MatrixXd(const MatrixXd& other)
        void setZero()
        void setZero(int, int)
        float& operator()(int, int) # read access

    # a vector
    cdef cppclass VectorXd:
        Eigen::VectorXd()
        VectorXd(int)
        VectorXd(const VectorXd& other)
        int rows()
        int cols()
        double& operator[](int)
        void setZero()
        void setZero(int)

        float& operator()(int) # read access



import numpy as np
import sympy as sp
from sympy import Matrix, Array

from . import rbda, se3

from ._version import get_versions


__version__ = get_versions()['version']
del get_versions


def ensure_positive_Z(u, vt, u_based_decision=True):
    '''
    This function is under the background of svd optimization. u or
    vt is understood as the normal vector of best fit plane of
    multi-sample points in 3d world.

    Parameters
    ----------
    u : ndarray
        u and vt are the output of 'linalg.svd' with matching inner
        dimensions so one can compute u @ vt
    vt: ndarray
        u and vt are the output of 'linalg.svd' with matching inner
        dimensions so one can compute u @ vt
    u_based_decision : bool, default=True
        If True, use the columns of u as the basis for sign flipping
        Otherwise, use the rows of v. The choice of which one is
        generally algorithm dependent.

    Returns
    -------
    u, vt : arrays with the same dimensions as the input.
    '''
    if u_based_decision:
        # sign is determined by Z location/index 2 of columns of u
        sign = np.sign(u[-1][2])
    else:
        # sign is determined by Z location/index 2 of rows of vt
        sign = np.sign(vt[-1, :][2])

    # Convert u to positive if it's negative, stay positive if not
    u[:, -1] *= sign
    # Make sure u and vt sign change together so that matrix multiplication
    # result won't change.
    vt[-1, :] *= sign
    return u, vt


def svd_without_sign_ambiguity(a):
    '''
    SVD with ensure deterministic.

    The raw svd function of numpy only ensure the multiplication result of
    column of u and matching row of vt is deterministic. If there is a matrix
    a in shape (M, N), both pair of (u[:, k], vt[k, :]) and pair of (-u[:,k],
    -vt[k, :]) are possible returned pair, where k = range(K), K = min(M, N).
    The sign ambiguity could exist for a pair of column u and matching a row
    of vt.
    The sign of each columns of u and each rows of vt are determined by a
    measure of left singular vector(u) with X and a measure of right singular
    vector(vt) with X. If they disagree with each other, the one with bigger
    abs value will be choose as determining factor, hence need to be always
    positive, and the other one will be flipped coorespondingly.
    This function implicitly use 'full_matrices=False'

    Parameters
    ----------
    a : (M, N) array
        A numpy 2D matrix, which should be legal for SVD. The legal check is
        handled by 'numpy.linalg.svd' inside function.

    Returns
    -------
    u : (M, K) array
        Left singular vectors. K = min(M, N)
    s: (, K) array
        Vector with the singular values.
    vt: (K, N) array
        Right singular vectors.

    References
    ----------
    .. [1] Bro, R., Acar, E., & Kolda, T. G. (2008). Resolving
           the sign ambiguity in the singular value decomposition
           Journal of Chemometrics: A Journal of the Chemometrics
           Society, 22(2), 135-140.
    .. [2] https://prod-ng.sandia.gov/techlib-noauth/access-control.cqi/2007/076422.pdf
    '''
    # svd dimensions:
    # u, s, vt = np.linalg.svd(X, full_matrices=False)
    # a = u @ diag(s) @ vt
    # (M, N) = (M, K) @ (K, K) @ (K, N)

    u, s, vt = np.linalg.svd(a, full_matrices=False)

    M = u.shape[0]
    N = vt.shape[1]
    K = s.shape[0]

    assert u.shape == (M, K)
    assert vt.shape == (K, N)
    assert a.shape == (M, N)

    sign_dict = {'left': np.zeros(K),
                 'right': np.zeros(K)}

    for k in range(K):
        mask = np.ones(K).astype(bool)
        mask[k] = False
        # (M, N) = (M, N) (M, K-1) @ (K-1, K-1) @ (K-1, N)
        Y = a - (u[:, mask] @ np.diag(s[mask]) @ vt[mask, :])
        for j in range(N):
            d = np.dot(u[:, k], Y[:, j])
            sign_dict['left'][k] += np.sum(np.sign(d) * d**2)
        for i in range(M) :
            d = np.dot(vt[k, :], Y[i, :])
            sign_dict['right'][k] += np.sum(np.sign(d) * d**2)

    for k in range(K):
        sign = 1.0
        if sign_dict['left'][k] < 0 and sign_dict['right'][k] < 0:
            sign = -1.0
        elif sign_dict['left'][k] * sign_dict['right'][k] < 0:
            if abs(sign_dict['left'][k]) < abs(sign_dict['right'][k]):
                # Use sign of right as determining flip sign.
                sign = np.sign(sign_dict['right'][k])
            else:
                sign = np.sign(sign_dict['left'][k])
        # multiplied by sign of itself will always change it to positive
        u[:, k] = u[:, k] * sign
        # alwars flip together.
        vt[k, :] = vt[k, :] * sign
    return u, s, vt


def svd_regulated(a):
    '''
    Wrapper function takes care of the replacement of standard svd
    to achieve goals below.
    1. sign of deterministic of svd
    2. In 3D world, the optimized normal vector achieved by svd is
    roughly aligned with Z-ish axis.

    The first goal is much more important in autodiff_svd. The second
    make sure normal vector only change in positive Z semi-space in
    3d. This make sure normal vector won't jog. This function implicityly
    return with 'full_matrices=False'
    '''
    u, s, vt = svd_without_sign_ambiguity(a)
    if a.shape[0] < a.shape[1]:  # bus
        u, vt = ensure_positive_Z(u, vt)
    else:  # rocket
        u, vt = ensure_positive_Z(u, vt, u_based_decision=False)
    return u, s, vt


def autodiff_svd(A, q_star, *, symbolic=False):
    '''
    Automatic Differentiating of svd.

    It takes matrix A, return differentiation of U, S and V with respect to
    specific q value q_star. U, S and VT(V.T) are result of standard SVD.
    This method is implemented in reverse mode in terms of auto diff conventions
    which is more effeciency since all free variables q take advantages of same
    intermedia svd result, U_star, S_star and V_star. It returns ecomonic result
    which means 'full_matrices=False' in corresponding 'numpy.linalg.svd'

    Parameters
    ----------
    A : (m, n) Matrix
        A sympy symbolic 2D Matrix, which should be legal for SVD later.
    q_star : dict
        A dict from sympy symbolic variable to value.
    symbolic : bool, optional
        Control return type. Return numerical result, numpy.ndarray, if true
        (default), otherwise return symbolic result, sympy.Array.

    Returns
    -------
    du : (Q, M, K) array
        Partial derivatives/gradient of U with respect to each variables in q star.
        Each variable matches a (M, K) matrix. Q is number of variables. Return
        sympy.Array if symbolic=True.
    ds : (0, K, K) array
        Partial derivatives/gradient of S with respect to each variables in q star,
        Each variable matches a (M, K) matrix. K = min(M, N). (..., K, K) is a strict
        diagonal matrix which is quaranteed by Hadamard operation. check examples
        for details. Return 'sympy.Array' if symbolic=True.
    dv : (Q, N, K) array
        Partial derivatives/gradient of V(NOT VT) with respect to each variables in
        q_star. Each variable matches a (N, K) matrix. Return 'sympy.Array' if
        symbolic=True

    See Also
    --------
    numpy.linalg.svd : Similar function in numpy.
    sympy.Array : 3D tensor in sympy.
    sympy.Matrix.jacobian : jacobian calculation in sympy.

    References
    ----------
    .. [1] James Townsend. Differentiating the Singular Value Decomposition.
    .. [2] Mike Giles. An extended collection of matrix derivative results
       for forward and reverse mode algorithmic differentiation.

    Examples
    --------
    >>> from sympy import sin, cos, symbols, Matrix
    >>> q0, q1 = symbols ("q0, q1")
    ... q star = (q0: 1, q1: 2}
    ... A = Matrix([[2*q0,          1,    q0**3 + 2*q1, sin(q0) + q1],
                   [q0**2 - q1**2,  2*q1, 2*q0 + q1**3, cos(q0) + q1],
                   [q0**2 + q1**2,  2,    3,            sin(q0) + cos(q1)]])
    >>> du, ds, dv = autodiff svd(A, q_star)
    >>> du.shape
    (2, 3, 3)
    >>> ds.shape
    (2, 3, 3)
    >>>dv.shape
    (2, 4, 3)
    >>> du_sym, ds_sym, dv_sym = autodiff_svd(A, q_star, symbolic=True)
    '''
    M, N = A.shape
    K = min(M, N)
    Q = len(q_star)
    # For a specific q0 and q1, get U_star, S_star and V_star_T.
    A_star = np.array(A.subs(q_star), dtype=float)
    U_star, S_star_1d, V_star_T = svd_regulated(A_star)  # U star (m, k), S star 1d (k, 1),
    U_star = U_star[:, :K]  # Use economic U (m, k)
    V_star_T = V_star_T[:K, :]  # Use economic VT (k, n)
    S_star = np.diag(S_star_1d)  # Construct S (k, k)
    S_star_inv = np.diag(np.reciprocal(S_star_1d))
    Ik = sp.eye(K)
    U_star_T = U_star.T
    V_star = V_star_T.T

    # F
    F_star = Matrix([[0 if i == j else 1 / (S_star_1d[j]**2 - S_star_1d[i]**2)
                    for j in range(K)]
                    for i in range(K)])

    # dA
    # Reshape A to 1d vector for convenience of jacobian calculation, will reshape back to 2
    A_element_wise = A.reshape(M * N, 1)

    q_var = Matrix(list(q_star.keys()))  # (g 0, q 1, qi. .., q Q-1]
    dA_element_wise = A_element_wise.jacobian(q_var)  # in shape (m*n, Q)
    # Reserve list of dA, d5, du and dv for each q variables. E.X. dA lsti] is dA with resp
    dA_lst = [None] * Q
    ds_lst = [None] * Q
    du_lst = [None] * Q
    dv_lst = [None] * Q

    Im = sp.eye(M)
    In = sp.eye(N)
    for i in range(Q):
        # Get matrix dA of eq(6) wrt qi, each element in dA lst is dA wrt qi. dA wrt gi is
        dA_lst[i] = dA_element_wise.T[i, :].reshape(M, N)
        # Use @ instead of * force all binary operation between np.array(s) and sp.Matrix(s)
        du_lst[i] = U_star @ F_star.multiply_elementwise(U_star_T @ dA_lst[i] @ V_star @ S_star
                                                         + S_star @ V_star_T @ dA_lst[i].T @ U_star) \
            + (Im - U_star @ U_star_T) @ dA_lst[i] @ V_star @ S_star_inv
        ds_lst[i] = Ik.multiply_elementwise(U_star_T @ dA_lst[i] @ V_star)
        dv_lst[i] = V_star @ F_star.multiply_elementwise(S_star @ U_star_T @ dA_lst[i] @ V_star
                                                         + V_star_T @ dA_lst[i].T @ U_star @ S_star) \
            + (In - V_star @ V_star_T) @ dA_lst[i].T @ U_star @ S_star_inv
    # Build a 3D tensor for all q variables
    du_tensor = Array([e.tolist() for e in du_lst])  # (Q, M, K)
    ds_tensor = Array([e.tolist() for e in ds_lst])  # (Q, K, K)
    dv_tensor = Array([e.tolist() for e in dv_lst])  # (Q, N, K)

    if symbolic:
        return du_tensor, ds_tensor, dv_tensor
    else:
        return (np.array(du_tensor.subs(q_star), dtype=float),
                np.array(ds_tensor.subs(q_star), dtype=float),
                np.array(dv_tensor.subs(q_star), dtype=float))



# Copyright (c) 2014,2015 Fabien Georget <fabieng@princeton.edu>, Princeton University
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#    * Neither the name of SpecMiCP nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



# Includes from Eigen
#
# This is kept simple on purpose
cdef extern from "<Eigen/Core>" namespace "Eigen":
    cdef cppclass Matrix[T, U, V]:
        Matrix()
        T& operator()(int, int)
        T& operator()(int)

    # A matrix
    cdef cppclass MatrixXd:
        MatrixXd()
        MatrixXd(int, int)
        MatrixXd(const MatrixXd& other)
        void setZero()
        void setZero(int, int)
        float& operator()(int, int) # read access

    # a vector
    cdef cppclass VectorXd:
        VectorXd()
        VectorXd(int)
        int rows() except +
        int cols()
        double& operator[](int) except +
        VectorXd(const VectorXd& other)
        void setZero()
        void setZero(int)
        double& operator()(int) # read access



from math import sin, cos
import numpy as np


def Rx(q):
    return np.array([[1,    0,          0],
                     [0,    cos(q), -sin(q)],
                     [0,    sin(q), cos(q)]])


def Ry(q):
    return np.array([[cos(q),   0,  sin(q)],
                     [0,        1,  0],
                     [-sin(q),  0,  cos(q)]])


def Rz(q):
    return np.array([[cos(q),   -sin(q),    0],
                     [sin(q),   cos(q),     0],
                     [0,        0,          1]])
